#!/usr/bin/env python
#
# -*- mode:python; sh-basic-offset:4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim:set tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8:
#

from clustoec2.drivers.categories.securitygroup import *



#!/usr/bin/env python
#
# -*- mode:python; sh-basic-offset:4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim:set tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8:
#

import os
import pprint
import sys

JSON = False
YAML = False
try:
    import yaml
    assert yaml
    YAML = True
except ImportError:
    pass

try:
    import simplejson as json
    assert json
    JSON = True
except ImportError:
    try:
        import json
        assert json
        JSON = True
    except:
        pass

import clusto
from clusto import script_helper
from clustoec2 import drivers as ec2_drivers


class Ec2(script_helper.Script):

    _instance_driver = ec2_drivers.servers.EC2VirtualServer
    _default_conn_manager = 'ec2connman'

    def __init__(self, *args, **kwargs):
        script_helper.Script.__init__(self, *args, **kwargs)
        self.formatters = {
            'pprint': (
                pprint.pformat, {'indent': 4},
            )
        }
        if YAML:
            self.formatters['yaml'] = (
                yaml.safe_dump,
                {
                    'encoding': 'utf-8',
                    'explicit_start': True,
                    'default_flow_style': False,
                },
            )
        if JSON:
            self.formatters['json'] = (
                json.dumps,
                {
                    'indent': 2,
                    'sort_keys': True
                }
            )

    def run(self, args):
        "Main run method"

        objects = []
        for _ in args.instances:
            try:
                objects.append(clusto.get_by_name(_))
            except LookupError:
                self.warn('Could not find instance %s in the clusto database' % (_,))
                if args.command == 'create':
                    self.info('Creating clusto object for %s' % (_,))
                    objects.append(self._instance_driver(_))
            except Exception as e:
                self.critical(e)
                return
        kwargs = dict(args.__dict__.items())
        for _ in ('command', 'config', 'dsn', 'loglevel', 'instances',):
            kwargs.pop(_)
        kwargs['objects'] = objects
        self.debug(kwargs)
        # Only the create command should (and in fact, *must*) receive an empty list of objects
        if not objects:
            self.error('Cannot run with an empty list of instances')
            return 1
        return (getattr(self, 'run_%s' % (args.command, ))(**kwargs))

    def _get_instance_data(self, instance):
        """
        Returns AWS data given an instance
        """
        connman = instance.attr_value(key='awsconnection', subkey='manager')
        data = connman._instance_to_dict(instance._instance)
        return data

    def run_show(self, **kwargs):
        "Prints the AWS data of the given objects to stdout"

        objs = []
        for obj in kwargs.get('objects'):
            objs.append({obj.name: self._get_instance_data(obj)})
        self.debug(objs)
        cb = self.formatters[kwargs.get('format', 'pprint')]
        print cb[0](objs, **cb[1])

    def run_state(self, **kwargs):
        "Prints the AWS state of the given objects to stdout"

        objs = []
        for obj in kwargs.get('objects'):
            objs.append({obj.name: obj.state})
        self.debug(objs)
        cb = self.formatters[kwargs.get('format', 'pprint')]
        print cb[0](objs, **cb[1])

    def _change_state(self, **kwargs):
        "Handles changing the state to running/stopped"

        state = kwargs['state']
        assert state in ('running', 'stopped',)
        wait = kwargs.get('wait', False)
        objs = kwargs.get('objects', [])
        confirm = False
        self.debug(objs)
        for obj in objs:
            self.info('Changing %s to %s state' % (obj.name, state, ))
            if state == 'running':
                confirm = obj.power_on()
            elif state == 'stopped':
                confirm = obj.power_off()
            else:
                pass
        # If waiting, only wait for the last one to change state
        if confirm and wait:
            self.info('Wait until instance(s) is/are in %s state' % (state,))
            obj.poll_until(state)

    def run_start(self, **kwargs):
        "Marshall over to _change_state"
        return self._change_state(state='running', **kwargs)

    def run_stop(self, **kwargs):
        "Marshall over to _change_state"
        return self._change_state(state='stopped', **kwargs)

    def run_create(self, **kwargs):
        "Create one or more EC2 instance(s) (if not exist)"
        objs = kwargs.get('objects', [])
        for obj in objs:
            try:
                self.info('Attempting to create %s' % (obj.name,))
                obj.create()
            except Exception as e:
                self.error('Error creating %s: %s' % (obj.name, e,))
        self.info('All objects created')
        return

    def _add_common_arguments(self, parser):
        parser.add_argument(
            '-k', '--aws-key', required=not os.environ.get('AWS_ACCESS_KEY_ID', False),
            help='Your AWS key id, defaults to ENV[AWS_ACCESS_KEY_ID] if set',
            default=os.environ.get('AWS_ACCESS_KEY_ID')
        )
        parser.add_argument(
            '-s', '--aws-secret-key', required=not os.environ.get('AWS_SECRET_ACCESS_KEY', False),
            help='Your AWS key secret, defaults to ENV[AWS_SECRET_ACCESS_KEY] if set',
            default=os.environ.get('AWS_SECRET_ACCESS_KEY')
        )
        parser.add_argument(
            '--conn-manager', '-c', default=self._default_conn_manager,
            help='Name of the EC2 Connection Manager you want to use'
        )
        formats = ['pprint']
        if YAML:
            formats.append('yaml')
        if JSON:
            formats.append('json')
        cmds = (
            'state',
            'show',
            'start',
            'stop',
            'create',
        )
        parser.add_argument(
            '-f', '--format', choices=formats, default='pprint',
            help='Default format to present results in'
        )
        parser.add_argument(
            '--wait', action='store_true', default=False,
            help='Wait for interaction with instances to finish (start/stop/create/destroy)'
        )
        parser.add_argument(
            '-p', '--pool', action='append', default=[],
            help='Add this instance to these pools before creating'
        )
        parser.add_argument(
            '-sg', '--security-group', action='append', default=[],
            help='Add new instance(s) to security group(s) (for create only). '
            'Keep in mind that security groups will be created if they don\'t exist'
        )
        parser.add_argument(
            'command', choices=cmds,
            help='EC2 command to run'
        )
        parser.add_argument(
            'instances', nargs='+', metavar='instance',
            help='EC2 instance(s) to interact with'
        )

    def _add_arguments(self, parser):
        self._add_common_arguments(parser)


def main():
    ec2, args = script_helper.init_arguments(Ec2)
    return(ec2.run(args))

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
#
# -*- mode:python; sh-basic-offset:4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim:set tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8:
#

from clusto.drivers.locations.zones import BasicZone
from clustoec2.drivers.base import VPCMixin


class VPCSubnet(BasicZone, VPCMixin):
    """
    VPC subnet driver.
    """

    _driver_name = "vpcsubnet"

    def __init__(self, name_driver_entity, **kwargs):
        super(VPCSubnet, self).__init__(name_driver_entity, **kwargs)

        self.set_attr(
            key='aws', subkey='ec2_subnet_id', value=kwargs.get(
                'subnet', name_driver_entity
            )
        )

    def _get_subnet(self):
        """
        Returns a boto.vpc.subnet.Subnet object to work with
        """

        return self._get_object('subnet')

    def get_cidr_block(self):
        return self._get_subnet().cidr_block

    _subnet = property(lambda self: self._get_subnet())
    state = property(lambda self: self._get_state('subnet'))
    cidr_block = property(lambda self: self.get_cidr_block())



#!/usr/bin/env python
#
# -*- mode:python; sh-basic-offset:4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim:set tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8:
#


from clusto.drivers import pool


class EC2SecurityGroup(pool.Pool):

    _driver_name = 'ec2_security_group'

    def __init__(self, name_driver_entity, **kwargs):
        pool.Pool.__init__(self, name_driver_entity, **kwargs)
        if 'group_id' in kwargs:
            self.set_attr(key='aws', subkey='ec2_security_group_id', value=kwargs.get('group_id'))
        if 'group_name' in kwargs:
            self.set_attr(key='aws', subkey='ec2_security_group', value=kwargs.get('group_name'))



#!/usr/bin/env python
#
# -*- mode:python; sh-basic-offset:4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim:set tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8:
#

from boto import ec2
from boto import vpc
import logging

from clusto.exceptions import ResourceException
from clustoec2.drivers.resourcemanagers import ec2connmanager


class VPCConnManagerException(ResourceException):
    pass


class VPCConnectionManager(ec2connmanager.EC2ConnectionManager):

    _driver_name = 'vpcconnmanager'
    _conns = {}

    def _connection(self, region=None):
        """
        Returns a connection "pool" (just a dict with an object per region
        used) to the calling code
        """
        r = region or 'us-east-1'
        if r not in self._conns:
            self._conns[r] = vpc.connect_to_region(
                r,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
            )
        return self._conns[r]

    def _instance_to_dict(self, instance):
        """
        Returns a dictionary with Instance information
        """
        return {
            'placement': instance.placement,
            'instance_id': instance.id,
            'image_id': instance.image_id,
            'region': instance.region.name,
            'subnet_id': instance.subnet_id,
            'vpc_id': instance.vpc_id,
        }

    def _subnet_to_dict(self, subnet):
        """
        Returns a dictionary with Instance information
        """
        return {
            'region': subnet.region.name,
            'availability_zone': subnet.availability_zone,
            'vpc_id': subnet.vpc_id,
            'subnet_id': subnet.id,
        }

    def _vpc_to_dict(self, ovpc):
        """
        Returns a dictionary with Instance information
        """
        return {
            'region': ovpc.region.name,
            'vpc_id': ovpc.id,
        }

    def additional_attrs(self, thing, resource, number=True):
        """
        Record the image allocation as additional resource attrs
        """

        for name, val in resource.items():
            if isinstance(val, vpc.subnet.Subnet):
                data = self._subnet_to_dict(val)
            elif isinstance(val, vpc.vpc.VPC):
                data = self._vpc_to_dict(val)
            elif isinstance(val, ec2.instance.Instance):
                data = self._instance_to_dict(val)
            else:
                data = None

            logging.debug(data)
            if data:
                thing.add_attr(
                    key=self._attr_name,
                    subkey=name,
                    number=number,
                    value=data
                )
                return data



#!/usr/bin/env python
#
# -*- mode:python; sh-basic-offset:4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim:set tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8:
#

from clustoec2.drivers.base import VPCMixin
from clustoec2.drivers.devices.servers import ec2server


class VPCVirtualServer(ec2server.EC2VirtualServer, VPCMixin):

    _driver_name = 'vpcvirtualserver'

    _instance = property(lambda self: self._get_instance())
    state = property(lambda self: self._get_instance_state())
    private_ips = property(lambda self: self.get_private_ips())



#!/usr/bin/env python
#
# -*- mode:python; sh-basic-offset:4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim:set tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8:
#

import os
import sys

import clusto
from clusto import script_helper
from clusto import drivers
from clustoec2 import drivers as ec2_drivers


class BootstrapEc2(script_helper.Script):
    """
    Will bootstrap your ec2 infrastructure (regions, zones, etc)
    """

    def __init__(self):
        script_helper.Script.__init__(self)

    def _add_arguments(self, parser):
        parser.add_argument(
            '-k', '--aws-key', required=not os.environ.get('AWS_ACCESS_KEY_ID', False),
            help='Your AWS key id, defaults to ENV[AWS_ACCESS_KEY_ID] if set',
            default=os.environ.get('AWS_ACCESS_KEY_ID')
        )
        parser.add_argument(
            '-s', '--aws-secret-key', required=not os.environ.get('AWS_SECRET_ACCESS_KEY', False),
            help='Your AWS key secret, defaults to ENV[AWS_SECRET_ACCESS_KEY] if set',
            default=os.environ.get('AWS_SECRET_ACCESS_KEY')
        )
        parser.add_argument(
            '--vpc-manager', '-V', default='vpcconnman',
            help='Name of the VPC Connection Manager you want to use'
        )
        parser.add_argument(
            '--conn-manager', '-c', default='ec2connman',
            help='Name of the EC2 Connection Manager you want to use'
        )
        parser.add_argument(
            '--add-to-pool', '-p', default=None,
            help='If given, amazon "location" objects will be inserted in '
            'the given pool'
        )
        parser.add_argument(
            '--no-import', default=False, action='store_true',
            help='Skip importing existing resources'
        )

    def add_subparser(self, subparsers):
        parser = self._setup_subparser(subparsers)
        self._add_arguments(parser)

    def run(self, args):
        self.debug('Grab or create the VM Manager')
        try:
            ec2connman = clusto.get_by_name(
                args.conn_manager,
                assert_driver=ec2_drivers.resourcemanagers.EC2ConnectionManager
            )
            vpcman = clusto.get_by_name(
                args.vpc_manager,
                assert_driver=ec2_drivers.resourcemanagers.VPCConnectionManager
            )
        except LookupError:
            if not args.aws_key and not args.aws_secrets_key:
                raise Exception(
                    'you must specify both an aws_access_key_id '
                    'and an aws_secret_access_key if you don\'t already have '
                    'an EC2ConnectionManager'
                )
            ec2connman = clusto.get_or_create(
                args.conn_manager,
                ec2_drivers.resourcemanagers.EC2ConnectionManager,
                aws_access_key_id=args.aws_key,
                aws_secret_access_key=args.aws_secret_key
            )
            self.info(
                'Created the "%s" EC2 Connection Manager' % (
                    args.conn_manager
                )
            )
            vpcman = ec2_drivers.resourcemanagers.VPCConnectionManager(
                args.vpc_manager,
                aws_access_key_id=args.aws_key,
                aws_secret_access_key=args.aws_secret_key
            )
            self.info(
                'Created the "%s" VPC Connection Manager' % (
                    args.vpc_manager
                )
            )

        container_pool = None
        if args.add_to_pool:
            container_pool = clusto.get_or_create(
                args.add_to_pool, drivers.pool.Pool)
        self.info('Creating all available regions')
        for region in ec2connman._connection().get_all_regions():
            curconn = ec2connman._connection(region.name)
            vpcconn = vpcman._connection(region.name)
            region_entity = clusto.get_or_create(
                region.name,
                ec2_drivers.locations.datacenters.EC2Region,
                region=region.name
            )
            region_entity.set_attr(
                key='aws', subkey='ec2_region',
                value=region.name
            )
            self.debug('Created "%s" region' % (region.name, ))
#           Create all VPCs (if any)
            self.info(
                'Creating all VPCs (if any) for region %s' % (
                    region.name,
                )
            )
            for v in vpcconn.get_all_vpcs():
                v_entity = clusto.get_or_create(
                    v.id,
                    ec2_drivers.locations.datacenters.VPC,
                    vpc=v.id,
                )
                if v_entity not in vpcman.referencers():
                    vpcman.allocate(v_entity)
                    vpcman.additional_attrs(
                        v_entity,
                        resource={'vpc': v}
                    )
                self.debug('Created "%s" VPC' % (v.id, ))
                if v_entity not in region_entity:
                    region_entity.insert(v_entity)
#               Create all subnets (if any)
                self.info(
                    'Creating all subnets (if any) for VPC %s' % (
                        v.id,
                    )
                )
                for sn in vpcconn.get_all_subnets(filters={'vpc_id': v.id}):
                    sn_entity = clusto.get_or_create(
                        sn.id,
                        ec2_drivers.locations.zones.VPCSubnet,
                        subnet=sn.id,
                    )
                    if sn_entity not in vpcman.referencers():
                        vpcman.allocate(sn_entity)
                        vpcman.additional_attrs(
                            sn_entity,
                            resource={'subnet': sn}
                        )
                    if sn_entity not in v_entity:
                        v_entity.insert(sn_entity)
                        self.debug('Inserted subnet %s in VPC %s' % (
                            sn.id, v.id, )
                        )
#           Create all zones
            self.info(
                'Creating all availability zones for region %s' % (
                    region.name,
                )
            )
            for zone in curconn.get_all_zones():
                zone_entity = clusto.get_or_create(
                    zone.name,
                    ec2_drivers.locations.zones.EC2Zone,
                    placement=zone.name
                )
                zone_entity.set_attr(
                    key='aws', subkey='ec2_placement',
                    value=zone.name
                )
                self.debug('Created "%s" zone' % (zone.name, ))
                if zone_entity not in region_entity:
                    region_entity.insert(zone_entity)
                self.debug(
                    'Inserted "%s" zone in "%s" region' % (
                        zone.name, region.name,
                    )
                )
#               if there are subnets in this region, insert them
                for sn in vpcconn.get_all_subnets(
                    filters={
                        'availability-zone': zone.name,
                    }
                ):
                    sn_entity = clusto.get_by_name(sn.id)
                    if sn_entity not in zone_entity:
                        zone_entity.insert(sn_entity)
                        self.debug('Inserted %s subnet in %s AZ' % (
                            sn.id, zone.name, )
                        )
            if container_pool and region_entity not in container_pool:
                self.debug(
                    'Adding region %s to pool %s' % (
                        region.name, args.add_to_pool,
                    )
                )
                container_pool.insert(region_entity)

        self.info('Creating all security groups')
        sgs = vpcman._connection().get_all_security_groups()
        belong = {}
        for sg in sgs:
            belong[sg.id] = [_.id for _ in sg.instances()]
        sgs = [(_.id, _.name, _.region.name, _.vpc_id) for _ in sgs]
        for sg_id, sg_name, region_name, vpc_id in sgs:
            self.debug(
                'Importing %s (%s), region: %s, vpc? %s' % (
                    sg_name, sg_id, region_name, bool(vpc_id),
                )
            )
            sg_ent = clusto.get_or_create(
                sg_id,
                ec2_drivers.categories.securitygroup.EC2SecurityGroup,
                group_id=sg_id,
                group_name=sg_name
            )
            if vpc_id:
                parent = clusto.get_by_name(vpc_id)
            else:
                parent = clusto.get_by_name(region_name)
            if sg_ent not in parent:
                self.debug('Inserting security group %s into %s' % (sg_id, vpc_id or region_name,))
                parent.insert(sg_ent)

        if not args.no_import:
            self.info('Creating all instances')
            for reservations in vpcman._connection().get_all_instances():
                for instance in reservations.instances:
                    idriver = ec2_drivers.devices.servers.EC2VirtualServer
                    connman = ec2connman
                    name = instance.tags.get('Name', instance.id).lower().replace(' ', '_')
                    if instance.vpc_id and instance.subnet_id:
                        idriver = ec2_drivers.devices.servers.VPCVirtualServer
                        connman = vpcman
                    self.debug('Creating %s instance (%s)' % (name, idriver, ))
                    instance_entity = clusto.get_or_create(
                        name,
                        idriver,
                    )
                    placement = clusto.get_by_name(instance.subnet_id or instance.placement)
                    self.debug('Inserting instance %s into %s' % (name, placement, ))
                    if instance_entity not in placement:
                        placement.insert(instance_entity)
                    instance_entity.set_attr(
                        key='aws', subkey='ec2_instance_type',
                        value=instance.instance_type
                    )
                    if instance.key_name is not None:
                        instance_entity.set_attr(
                            key='aws', subkey='ec2_key_name',
                            value=instance.key_name
                        )
                    instance_entity.set_attr(
                        key='aws',
                        subkey='ec2_instance_id',
                        value=instance.id,
                    )

                    for sg, instances in belong.items():
                        if instance.id in instances:
                            sg_ent = clusto.get_by_name(sg)
                            if instance_entity not in sg_ent:
                                self.debug(
                                    'Adding instance %s to security group %s' % (
                                        instance.id, sg,
                                    )
                                )
                                sg_ent.insert(instance_entity)

                    self.debug('Allocating instance %s from %s' % (name, connman, ))
                    if instance_entity not in connman.referencers():
                        connman.allocate(instance_entity)
                        connman.additional_attrs(
                            instance_entity,
                            resource={'instance': instance}
                        )
                        instance_entity.update_metadata()
                    self.debug('%s is imported' % (instance,))
        self.info('Finished, AWS objects should now be in the database')


def main():
    bootstrap, args = script_helper.init_arguments(BootstrapEc2)
    return(bootstrap.run(args))

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
#
# -*- mode:python; sh-basic-offset:4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim:set tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8:
#

import setuptools


setuptools.setup(
    name='clusto-ec2',
    version='0.6.4',
    packages=setuptools.find_packages('src'),
    author='Ron Gorodetzky',
    author_email='ron@parktree.net',
    description='Amazon EC2 extension for clusto',
    install_requires=[
        'clusto>0.6',
        'boto>=2.0',
        'mako',
    ],
    entry_points={
        'console_scripts': [
            'clusto-ec2 = clustoec2.commands.ec2:main',
            'clusto-vpc = clustoec2.commands.vpc:main',
            'clusto-ec2-bootstrap = clustoec2.commands.bootstrap:main',
        ],
    },
    zip_safe=False,
    package_dir={
        '': 'src',
    },
)



#!/usr/bin/env python
#
# -*- mode:python; sh-basic-offset:4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim:set tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8:
#

from boto.ec2 import blockdevicemapping
from clusto.drivers.devices.servers import BasicVirtualServer
from clusto.exceptions import ResourceException
from clustoec2.drivers.base import EC2Mixin
from datetime import datetime
import IPy
from mako import template
import os
import time

MAX_POLL_COUNT = 30


class EC2VirtualServer(BasicVirtualServer, EC2Mixin):

    _driver_name = 'ec2virtualserver'
    _i = None
    _int_ip_const = 2147483648
    # non-listed instances will get 0 ephemeral drives at spin-up time
    _eph_drives = {
        'c1.medium': 1,
        'c1.xlarge': 4,
        'c3.large': 2,
        'c3.xlarge': 2,
        'c3.4xlarge': 2,
        'c3.8xlarge': 2,
        'cc2.8xlarge': 4,
        'cg1.4xlarge': 2,
        'cr1.8xlarge': 2,
        'd2.xlarge': 3,
        'd2.2xlarge': 6,
        'd2.4xlarge': 12,
        'd2.8xlarge': 24,
        'g2.2xlarge': 1,
        'g2.8xlarge': 2,
        'hi1.4xlarge': 2,
        'hs1.8xlarge': 24,
        'i2.xlarge': 1,
        'i2.2xlarge': 2,
        'i2.4xlarge': 4,
        'i2.8xlarge': 8,
        'm1.small': 1,
        'm1.medium': 1,
        'm1.large': 2,
        'm1.xlarge': 4,
        'm2.xlarge': 1,
        'm2.2xlarge': 1,
        'm2.4xlarge': 2,
        'm3.medium': 1,
        'm3.large': 1,
        'm3.xlarge': 2,
        'm3.2xlarge': 2,
        'r3.large': 1,
        'r3.xlarge': 1,
        'r3.2xlarge': 1,
        'r3.4xlarge': 1,
        'r3.8xlarge': 2,
    }

    def _int_to_ipy(self, num):
        return IPy.IP(num + self._int_ip_const)

    def _get_instance(self):
        """
        Returns a boto.ec2.Instance object to work with
        """

        reservation = self._get_object('instance')
        if reservation:
            return reservation.instances[0]

    def console(self, *args, **kwargs):
        """
        Returns the console log output from the EC2 instance
        """
        console = self._get_instance().get_console_output()
        return console.output

    def get_private_ips(self):
        """
        Only return private IP addresses
        """
        return self.get_ips(objects=True, private=True, public=False)

    def get_public_ips(self):
        """
        Only return public IP addresses
        """
        return self.get_ips(objects=True, private=False, public=True)

    def get_ips(self, objects=False, private=True, public=True):
        """
        Returns a list of IP addresses to work with. Alternatively,
        it can return a list of IPy.IP objects.
        """
        ips = []
        if private:
            l = self.attr_values(key='ip', subkey='nic-eth')
            if l:
                if objects:
                    [ips.append(self._int_to_ipy(_)) for _ in l]
                else:
                    [ips.append(self._int_to_ipy(_).strNormal()) for _ in l]
        if public:
            l = self.attr_values(key='ip', subkey='ext-eth')
            if l:
                if objects:
                    [ips.append(self._int_to_ipy(_)) for _ in l]
                else:
                    [ips.append(self._int_to_ipy(_).strNormal()) for _ in l]
        return ips

    def update_metadata(self, *args, **kwargs):
        """
        Updates the IP attributes for this instance
        """

        self.clear_metadata()
        self._get_instance().update()
        ip = self._get_instance().private_ip_address
        if ip:
            self.add_attr(
                key='ip',
                subkey='nic-eth',
                value=IPy.IP(ip).int() - self._int_ip_const,
                number=0
            )
            self.add_attr(
                key='ip',
                subkey='ipstring',
                value=ip,
                number=0
            )
        ip = self._get_instance().ip_address
        if ip:
            self.add_attr(
                key='ip',
                subkey='ext-eth',
                value=IPy.IP(ip).int() - self._int_ip_const,
                number=1
            )
            self.add_attr(
                key='ip',
                subkey='ipstring',
                value=ip,
                number=1
            )

    def clear_metadata(self, *args, **kwargs):
        """
        Deletes the metadata that can change
        """
        self.del_attrs(key='ip')

    def power_off(self, captcha=True):
        if captcha and not self._power_captcha('shutdown'):
            return False
        self._get_instance().stop()
        return True

    def power_on(self, captcha=False):
        if captcha and not self._power_captcha('start'):
            return False
        self._get_instance().start()
        return True

    def power_reboot(self, captcha=True):
        if captcha and not self._power_captcha('reboot'):
            return False
        self._get_instance().reboot()

    def _build_user_data(self, udata=None):
        """
        Builds and returns a userdata string based on the
        user data string attribute
        """

        udata = self.attr_value(
            key='aws', subkey='ec2_user_data',
            merge_container_attrs=True
        )

        if udata:
            tpl = template.Template(udata)
            # Always send the name of this object
            attr_dict = {
                'name': self.name,
            }
            # Add all aws information as values
            for attr in self.attrs(key='aws', merge_container_attrs=True):
                # don't recurse
                if attr.subkey == 'ec2_user_data':
                    continue
                if attr.subkey == 'ec2_boot_script_file':
                    if os.path.isfile(attr.value):
                        f = open(attr.value, 'rb')
                        attr_dict[attr.subkey] = f.read()
                        f.close()
                else:
                    attr_dict[attr.subkey] = attr.value
            attr_dict.update({'name': self.name, })
            return tpl.render(**attr_dict)
        else:
            return None

    def _ephemeral_storage(self, instance_type):
        """
        Return the appropriate block mapping so you
        get your ephemeral storage drives
        """

        number = self._eph_drives.get(instance_type, 0)
        mapping = blockdevicemapping.BlockDeviceMapping()
        for block in range(0, number):
            eph = blockdevicemapping.BlockDeviceType()
            eph.ephemeral_name = 'ephemeral%d' % (block, )
            mapping['/dev/sd%s' % (chr(ord('b') + block),)] = eph
        return mapping

    def _get_or_create_security_groups(self, conn, vpc_id=None):
        """
        If security groups don't exist, they will get created. Results will
        be returned to calling argument. It receives the current connection
        used as a parameter
        """

        # We gotta search for both, because if they don't exist you'll have to create them
        sgs_ids = self.attr_values(
            key='aws',
            subkey='ec2_security_group_id',
            merge_container_attrs=True
        )
        sgs_names = self.attr_values(
            key='aws',
            subkey='ec2_security_group',
            merge_container_attrs=True
        )
        filters = {}
        # If you received a vpc_id then only return those from that vpc
        # Also if you received a vpc_id, you must return ids
        ids = False
        if vpc_id:
            filters['vpc-id'] = vpc_id
            ids = True
        existing_groups = dict([(_.id, _.name) for _ in conn.get_all_security_groups(filters=filters)])
        final_groups = set()

        # If you have a security group id that doesn't exist in aws (???) bail out
        diff = set(sgs_ids) - set(existing_groups.keys())
        if diff:
            raise ValueError(
                'The security group ids %s are not present in the security groups '
                'found in amazon, where did you find them?' % (','.join(diff),)
            )

        # I need the two items otherwise I'd do this with sets
        for sg in sgs_ids:
            [final_groups.add((k, v)) for k, v in existing_groups.iteritems() if k == sg]

        # Next search based on the group name. It is possible group names don't exist
        # in aws (new group) so you have to create those.
        for sg in sgs_names:
            if sg not in existing_groups.values():
                desc = 'Created on %s' % (datetime.now(),)
                group = conn.create_security_group(name=sg, description=desc)
                final_groups.add((group.id, group.name))
            [final_groups.add((k, v)) for k, v in existing_groups.iteritems() if v == sg]

        final_groups = dict(final_groups)
        if ids:
            return final_groups.keys()
        else:
            return final_groups.values()

    def create(self, captcha=False, wait=True):
        """
        Creates an instance if it isn't already created
        """

        try:
            assert self._instance
        except AssertionError:
            pass
        except:
            raise('Cannot create this instance')

        res = self._mgr_driver.resources(self)[0]
        mgr = self._mgr_driver.get_resource_manager(res)

        # We build these on a different step
        skip_attrs = [
            'ec2_security_group',
            'ec2_security_group_id',
            'ec2_user_data'
        ]

        # Grab all the `ec2_*` attributes available
        ec2_attrs = dict(
            [
                (_.subkey, _.value) for _ in self.attrs(
                    key='aws', merge_container_attrs=True
                ) if _.subkey
                and _.subkey.startswith('ec2_')
                and _.subkey not in skip_attrs
            ]
        )

        image_id = ec2_attrs.pop('ec2_ami', None)
        if not image_id:
            raise ResourceException(
                'No image specified for %s' % (
                    self.name,
                )
            )

        region = ec2_attrs.pop('ec2_region', 'us-east-1')

        instance_type = ec2_attrs.pop('ec2_instance_type', None)
        if not instance_type:
            raise ResourceException(
                'No instance type specified for %s' % (
                    self.name,
                )
            )

        placement = ec2_attrs.pop('ec2_placement', None)
        user_data = self._build_user_data()
        key_name = ec2_attrs.pop('ec2_key_name', None)

        image = mgr._connection(region).get_image(image_id)
        # Unless you explicitly skip the creation of ephemeral drives, these
        # will get created, you're already paying for them after all
        block_mapping = None
        skip_ephemeral = ec2_attrs.pop('ec2_skip_ephemeral', False)
        if not skip_ephemeral:
            block_mapping = self._ephemeral_storage(instance_type)

        # Now we need to check if this is vpc or not
        vpc_id = self.attr_value(
            key='aws', subkey='vpc_id', merge_container_attrs=True, default=None
        )

        extra_args = dict(
            ('_'.join(_.split('_')[1:]), __) for _, __ in ec2_attrs.items()
        )

        if vpc_id:
            security_group_ids = self._get_or_create_security_groups(
                mgr._connection(region), vpc_id=vpc_id
            )
            extra_args['security_group_ids'] = security_group_ids
        else:
            security_groups = self._get_or_create_security_groups(
                mgr._connection(region)
            )
            extra_args['security_groups'] = security_groups

        reservation = image.run(
            instance_type=instance_type,
            placement=placement,
            key_name=key_name,
            user_data=user_data,
            block_device_map=block_mapping,
            **extra_args
        )

        self._i = reservation.instances[0]
        self._i.add_tag('Name', self.name)
        result = mgr.additional_attrs(self, resource={'instance': self._i}, number=res.number)
        if wait:
            self.poll_until('running')

        return (result, True)

    def poll_until(self, state, interval=2, max_poll=MAX_POLL_COUNT):
        """
        Polls for the requested status, with a possible timeout.
        Shamelessly stolen from py-smartdc
        """

        c = 0
        while self.state != state and c < max_poll:
            c += 1
            time.sleep(interval)

    def poll_while(self, state, interval=2, max_poll=MAX_POLL_COUNT):
        """
        Polls whil the status doesn't change, with a possible timeout.
        Shamelessly stolen from py-smartdc
        """

        c = 0
        while self.state == state and c < max_poll:
            c += 1
            time.sleep(interval)

    def destroy(self, captcha=True, wait=True):
        """
        Destroys this instance if it exists
        """
        if captcha and not self._power_captcha('destroy'):
            return False

        try:
            assert self._instance
        except AssertionError:
            raise ResourceException('This instance does not exist')
        except:
            pass

        instance_id = self._get_instance().id
        volumes = self._get_instance().connection.get_all_volumes(
            filters={'attachment.instance-id': instance_id})
        self._get_instance().terminate()
        if wait:
            self.poll_until('terminated')
        # destroy all volumes
        warnings = []
        for vol in volumes:
            dev = vol.attach_data.device.split('/')[-1]
            # It could be that some instances haven't freed their volumes
            try:
                vol.delete()
            except Exception as e:
                warnings.append(
                    'Could not delete volume %(dev)s (%(id)s) '
                    'from %(name)s, reason: %(reason)s' % {
                        'dev': dev,
                        'id': vol.id,
                        'name': self.name,
                        'reason': e.error_message
                    }
                )

        # finally, delete the entity from clustometa
        self.entity.delete()
        return warnings

    def reconcile_ebs_volumes(self):
        """
        Will reflect the changes from amazon in clusto first,
        whatever's left from clusto to amazon
        """

        volumes = {}
        conn = self._get_instance().connection
        # Seems important to grab the placement from the instance data in the
        # unlikely scenario the clusto data doesn't match?
        zone = self._get_instance().placement
        instance_id = self._get_instance().id
        for attr in self.attrs(key='aws', merge_container_attrs=True):
            if not attr.subkey.startswith('ebs_'):
                continue
            volume = '_'.join(attr.subkey.split('_')[1:])
            if volume not in volumes.keys():
                volumes[volume] = {}
            try:
                volumes[volume]['size'] = int(attr.value)
            except ValueError:
                if attr.value.startswith('vol-'):
                    volumes[volume]['vol-id'] = attr.value
                else:
                    volumes[volume]['extra'] = attr.value

        for dev, data in volumes.items():
            device = '/dev/%s' % (dev, )
            vol = None
            if 'vol-id' not in data.keys():
                # create the volumes w/ default size of 10G
                vol = conn.create_volume(int(data['size']), zone)
                self.add_attr(
                    key='aws',
                    subkey='ebs_%s' % (dev,), value=vol.id
                )
            else:
                try:
                    vol = conn.get_all_volumes(volume_ids=[data['vol-id']])
                    vol = vol[0]
                except:
                    # This volume does not exist anymore
                    self.del_attrs(key='aws', subkey='ebs_%s' % (dev,))
            if vol:
                # Attach the volume if it's not attached
                if not vol.attachment_state():
                    vol.attach(instance_id, device)
                else:
                    # Ok so it's attached, but what if it's attached to something
                    # else? if that is the case we should clear the attrs
                    try:
                        is_mine = conn.get_all_volumes(
                            volume_ids=[data['vol-id']],
                            filters={'attachment.instance-id': instance_id})
                    except:
                        is_mine = False
                    if is_mine:
                        vol.add_tag('Name', '%s:%s' % (self.name, device,))
                    else:
                        self.del_attrs(key='aws', subkey='ebs_%s' % (dev,))

        # Ok so now from aws to clusto
        volumes = conn.get_all_volumes(
            filters={'attachment.instance-id': instance_id})

        for vol in volumes:
            device = vol.attach_data.device
            dev = device.split('/')[-1]
            # update attrs that are not in our db
            if not self.attrs(key='aws', subkey='ebs_%s' % (dev,)):
                self.add_attr(
                    key='aws', subkey='ebs_%s' % (dev,),
                    value=int(vol.size)
                )
                self.add_attr(
                    key='aws', subkey='ebs_%s' % (dev,),
                    value=vol.id
                )
            tag = '%s:%s' % (self.name, device)
            if 'Name' not in vol.tags or vol.tags['Name'] != tag:
                vol.add_tag('Name', tag)

    def _get_instance_state(self):
        return self._get_instance().update()

    _instance = property(lambda self: self._get_instance())
    state = property(lambda self: self._get_instance_state())
    private_ips = property(lambda self: self.get_private_ips())
    public_ips = property(lambda self: self.get_public_ips())
